# -*- coding: utf-8 -*-
"""
Integration with django.

All django dependant code should live in this package. Everything else
should not depend on django in any way.
"""
from __future__ import absolute_import, unicode_literals
from .endpoint import DjangoEndpoint



# -*- coding: utf-8 -*-
"""
Rest endpoint base implementation.

Endpoints take care of interacting with HTTP. They for a bridge between HTTP
requests/responses and the REST resources. They should never be implemented
manually and ideally most apps should not need to subclass those but instead
implement all of their logic in the `RestResource` subclasses.

The code here is just an abstract base that holds the code common to all
endpoints. It is not self sufficient, but only minor web-framework dependent
code is left for the subclasses.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json
from logging import getLogger

# local imports
from .resource import RestResource
from . import filters


L = getLogger(__name__)


class RestEndpoint(object):
    """ Abstract base class for all REST endpoints.

    The endpoint responsibility is to provide a bridge between HTTP and
    `RestResource`.  The latter should never directly deal with HTTP.

    The subclasses must implement the following methods:

    - `is_http_response()`
    - `json_response()`

    """

    def __init__(self, res_cls):
        if not (
            isinstance(res_cls, type) and
            issubclass(res_cls, RestResource)
        ):
            raise ValueError('res_cls must be a subclass of RestResource')

        self.resource = res_cls()

    def get_ok_status(self, verb):
        """ Get a valid ok status code for a given rest verb. """
        if verb == 'create':
            return 201
        elif verb == 'delete':
            return 204
        else:
            return 200

    def dispatch(self, request, **keys):
        """ Dispatch the request to the appropriate resource method.

        This should handle both standard REST verbs as well as custom
        actions supported by the resource.

        :param HttpRequest request:
            The HTTP request.
        :param Any pk:
            The primary key for the object. This should be taken from the URL
            and if present it means the detail view was called. Lack of ``pk``.
            indicates that the request was made against the list URL.
        :return RestResponse:
            The ``RestResponse`` received from the resource implementation.
        """
        request.rest_keys = keys

        my_pk = request.rest_keys.get(self.resource.name + '_pk')

        rest_verb = self.determine_rest_verb(request.method, my_pk)
        if rest_verb is None:
            return self.json_response(405)

        handler = getattr(self.resource, rest_verb, None)

        if handler is None or not callable(handler):
            return self.json_response(405)

        try:
            handler_args = self.build_handler_args(rest_verb, request)
            result = handler(request, **handler_args)

            if self.is_http_response(result):
                return result
            else:
                return self.json_response(self.get_ok_status(rest_verb), result)

        except ValueError as ex:
            L.exception('Invalid REST invocation')
            L.error(ex)
            return self.json_response(400, {'detail': str(ex)})

        except NotImplementedError as ex:
            return self.json_response(405, {'detail': str(ex)})

        except Exception as ex:
            L.exception('Unhandled resource exception')
            L.error(ex)
            return self.json_response(500, {'detail': str(ex)})

    def is_http_response(self, result):
        """ Check if the given object is complete HTTP response.

        **Has to be implemented by subclasses.**

        If this method returns False, the *response* will be used to build the
        actual response that is returned to the calling endpoint. Having this
        in a separate method makes it possible to customise this behaviour in
        the subclasses and thus possibly support more web frameworks in the
        future.

        :param Any result:
            Resource handler result. This is what is returned by one of the
            handling methods of a given `RestResource`.
        :return:
            **True** if the response can be directly returned to the calling
            framework, **False** if it's just a data and the actual framework
            specific response has to still be built.
        """
        raise NotImplementedError(
            "RestEndpoint sublasses must implement .is_http_response()"
        )

    def json_response(self, status, data=None):
        """ Return a framework specific HTTP response.

        **Has to be implemented by subclasses.**

        :param int status:
            HTTP status code.
        :param dict|list data:
            Data that will be JSON serialized and sent as the response body.
        :return:
            Framework specific HTTP response object.
        """
        raise NotImplementedError(
            "RestEndpoint sublasses must implement .json_response()"
        )

    @classmethod
    def determine_rest_verb(self, http_method, pk):
        """ Return the REST operation associated with the given HTTP method.

        This is part of the `RestEndpoint` class because this way it becomes
        customisable by subclasses.

        :param str|unicode http_method:
            The HTTP method used for the request.
        :param str|int pk:
            The primary key extracted from the URL template.
        :return str|unicode:
            The REST verb associated with that request. This is also the name
            of the method on `RestResource` that will be called by the endpoint.
        """
        if http_method == 'GET':
            return 'query' if pk is None else 'get'
        elif http_method == 'POST' and pk is None:
            return 'create'
        elif http_method == 'PUT' and pk is not None:
            return 'update'
        elif http_method == 'DELETE' and pk is not None:
            return 'delete'
        elif http_method == 'OPTIONS':
            return 'options'
        elif http_method == 'HEAD':
            return 'head'
        else:
            return None

    @classmethod
    def build_handler_args(self, verb, request):
        """ Build handler invocation arguments.

        The parameters extracted here can be directly passed to the
        `RestResource` handler method for the given REST *verb*.

        This is part of the `RestEndpoint` class because this way it becomes
        customisable by subclasses.

        :param str|unicode verb:
            REST verb/action. This is what the `RestResource` classes abstract.
            Can be one of ``get/query/create/update/delete/options/head``.
        :param HttpRequest request:
            An HTTP request the handler will serve. Handler args are extracted
            from it here.
        :param None|int|str|unicode pk:
            The primary key of a resource that will be modified by request or
            ``None`` if the request is to list handlers. This is most likely
            extracted from the URL template and that's why it's not passed as
            part of the request.
        :return dict:
        """
        handler_args = {}

        if verb in ('create', 'update'):
            try:
                if request.body:
                    body = request.body.decode('utf-8')
                    handler_args['data'] = json.loads(body)
                else:
                    handler_args['data'] = None
            except:
                raise ValueError("Invalid JSON body: '{}'".format(
                    request.body.decode('utf-8')
                ))

        if verb in ('query',):
            handler_args['filters'] = filters.extract(request.GET)

        return handler_args



# -*- coding: utf-8 -*-
"""
REST resource base class.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import re


class RestResource(object):
    """ Represents the operations available on the given resource.

    This class should not handle HTTP directly. Rather than that there should
    be a separate implementation that maps HTTP layer onto resources. This way
    we can swap the HTTP layer without reimplementing the resource. This also
    follows SRP.

    If the given `RestResource` subclass doesn't have one of the REST methods
    implemented, the endpoint should return 405 Method not allowed (or 404?)
    """

    _NAME_RE = re.compile(r'^[a-zA-Z_][\w\d_]*$')
    name = None
    parent = None

    def __init__(self):
        self._validate_name(self.name)

    def _validate_name(self, name):
        """ Validate the name is a valid resource name. """
        if name is None or not self._NAME_RE.match(name):
            raise ValueError(
                'REST endpoint name contain only letters, digits and '
                'underscores and cannot start with a digit'
            )

    def query(self, request, filters):
        """ GET list. """
        raise NotImplementedError("query() not implemented by this resource")

    def get(self, request):
        """ GET detail. """
        raise NotImplementedError("get() not implemented by this resource")

    def create(self, request, data):
        """ POST list. """
        raise NotImplementedError("create() not implemented by this resource")

    def update(self, request, data):
        """ PUT detail. """
        raise NotImplementedError("update() not implemented by this resource")

    def delete(self, request):
        """ DELETE detail. """
        raise NotImplementedError("delete() not implemented by this resource")

    def options(self, request):
        """ OPTIONS list/detail. """
        raise NotImplementedError("options() not implemented by this resource")

    def head(self, request):
        """ OPTIONS list/detail. """
        raise NotImplementedError("options() not implemented by this resource")

    def get_my_pk(self, request):
        """ Read the current resource type PK from the request.

        :param HttpRequest request:
            An HTTP request. Must contain the ``.rest_keys`` dictionary injected
            by `RestEndpoint.dispatch()` method.
        :return:
            Primary key that is associated with this resource type. This is
            extracted from ``.rest_keys`` based on the resource name.
        """
        if hasattr(request, 'rest_keys'):
            return request.rest_keys.get(self.name + '_pk')



# -*- coding: utf-8 -*-
"""
Simple REST API implementation.

Each endpoint has it's resource class that represents all the operations
possible in that endpoint. The resource implementation should be framework
agnostic so it can be used both on AppEngine and regular django (possibly other
frameworks as well..
"""
from __future__ import absolute_import, unicode_literals
from .core.endpoint import RestEndpoint
from .core.resource import RestResource
from .core.routing import make_urls



# -*- coding: utf-8 -*-
"""
Helpers for setting up mapping between URLs and resources.
"""
from __future__ import absolute_import, unicode_literals
from .. import RestResource


def make_urls(default_endpoint_cls, endpoints):
    """

    :param list<RestEndpoint> endpoints:
    :return list<url>:

    """
    urls = []

    for endpoint in endpoints:
        if isinstance(endpoint, type) and issubclass(endpoint, RestResource):
            urls += default_endpoint_cls(endpoint).urls
        else:
            urls += endpoint.urls

    return urls



# -*- coding: utf-8 -*-
"""
Django implementation of a `RestEndpoint`.
"""
from __future__ import absolute_import, unicode_literals

# 3rd party imports
from django.conf.urls import url
from django.http import HttpResponse, JsonResponse

# local imports
from .. import RestEndpoint


class DjangoEndpoint(RestEndpoint):
    """ Endpoint implementation to use in django projects. """
    
    def __call__(self, *args, **kw):
        """ Django 'calls' views so this is what will be ran by django.

        This should cover the whole request lifecycle.
        """
        return self.dispatch(*args, **kw)

    def is_http_response(self, result):
        return isinstance(result, HttpResponse)

    def json_response(self, status, data=None):
        data = data or {}
        return JsonResponse(status=status, safe=False, data=data)

    @property
    def urls(self):
        """ Return all the URLs required by the endpoint.

        :return list<url>:
            A list of URLs that can be directly appended to ``urlpatterns`` of
            choice.
        """
        # TODO: Using resource.name for both prefix and PK name might blow up.
        # There might be one generic endpoint called /bars and we might want to
        # create a nested one /foo/bars. In those cases we have two resources
        # with the same name - bars.
        #
        # Another question is whether this is an issue. In the above example
        # the resources are not at the same 'level' in the tree. This will
        # definitely fail if we have a recursive endpoint, but that might be
        # even worse.
        name = self.resource.name
        ret = [
            url(
                '{name}/(?P<{name}_pk>[^/]+)/?$'.format(name=name),
                self.dispatch,
                name='{name}-detail'.format(name=name)
            ),
            url(
                '{name}/?$'.format(name=name),
                self.dispatch,
                name='{name}-list'.format(name=name)
            ),
        ]
        return ret
